#include "mex.h"
#include <iostream>
#include <stack>
#include <queue>
#include <vector>
#include <math.h>

/*  Pixel Class */
class pixel{
public:
    int x;
    int y;
    double gray;
    int label;
    
    
    pixel();
    pixel(int xi, int yi, double igray);
    pixel(int ix, int iy, int ilabel, double igray);
    ~pixel();
};

pixel::pixel()
{x=0;
 y=0;
 gray=0;
 label=0;
}
pixel::pixel(int xi, int yi, double igray)
{x=xi;
 y=yi;
 gray=igray;
 label=0;
}

pixel::~pixel() {//nothing
}


pixel::pixel(int ix, int iy, int ilabel, double igray)
{x=ix;
 y=iy;
 gray=igray;
 label=ilabel;
}


/* Couple struct */
struct couple {
    double k;     // Priority
    int label; // Value
    int x;
    int y;
    couple() : k(0), label(0) {}
    couple(double k, int t, int x, int y) : k(k), label(t), x(x), y(y) {}
};

/* Comparison Function on the Priority */
class C {
public:
    bool operator()(const couple &a1, const couple &a2) {
        return a1.k < a2.k ;
    }
};

using namespace std;
typedef stack<pixel> pix;
typedef stack<int> pi;
typedef vector<int> ve;
typedef vector<double> vd;
typedef vector<pixel> vi;

// Complementary functions

vi labelization(vi im_lab, int h, int w)//Label of the flat zone of the image
{
    int i, j, m, n;
    double t;
    int l=0;
    pix pile;
    pixel temp;
    
    for(j=0;j<h;j++)
        for(i=0;i<w;i++){
        if(im_lab[i+j*w].label==0){//flat zone detection
            t=im_lab[i+j*w].gray;
            l++;
            im_lab.at(i+j*w).label=l;
            for(m=j-1;m<j+2;m++)
                for(n=i-1;n<i+2;n++)
                    if((0<=m)&&(m<h)&&(0<=n)&&(n<w)&&(im_lab[n+w*m].gray==t)){
                pile.push(im_lab[n+w*m]);
                im_lab.at(n+w*m).label=l;
                    }
            while(!pile.empty()){
                temp=pile.top();
                pile.pop();
                for(m=temp.y-1;m<temp.y+2;m++)
                    for(n=temp.x-1;n<temp.x+2;n++)
                        if((0<=m)&&(m<h)&&(0<=n)&&(n<w)&&(im_lab[n+w*m].gray==t)&&(im_lab[n+w*m].label==0)){
                    pile.push(im_lab[n+w*m]);
                    im_lab[n+w*m].label=l;
                        }
            }
        }
        }
    return im_lab;
}

int find_max_label(vi im) {
    int i;
    int max=0;
    for(i=0;i<im.size();i++)
        if(max<im[i].label)
            max=im[i].label;
    return max;
}
ve compute_area(vi im, int nb_label) {
    int i;
    ve area_tab(nb_label);
    for(i=0;i<im.size();i++)
        area_tab.at(im[i].label-1)++;
    return area_tab;
}
vd compute_gray_level(vi im, int nb_label) {
    int i;
    vd gray_level_tab(nb_label);
    for(i=0;i<im.size();i++)
        gray_level_tab.at(im[i].label-1)=im[i].gray;
    return gray_level_tab;
}
int* find_min_area(ve area_tab) {
    int i;
    int* min;
    min=(int*)calloc(2, sizeof(int));
    min[0]=10000;
    
    for(i=0;i<area_tab.size();i++)
        if(min[0]>area_tab[i]) {
        min[0]=area_tab[i];
        min[1]=i+1;
        }
    return min;
}
int indice_max_value(vd tab) {
    int i, ind;
    double temp=0;
    ind=0;
    for(i=0;i<8;i++)
        if(temp<tab[i]) {
        temp=tab[i];
        ind=i;
        }
    return ind;
}

int* convert_indice_2_neighbors(int i) {
    int* NN;
    NN=(int*)calloc(2, sizeof(int));
    if(i==0) {
        NN[0]=-1;
        NN[1]=-1;
    }
    else if(i==1) {
        NN[0]=0;
        NN[1]=-1;
    }
    else if(i==2) {
        NN[0]=1;
        NN[1]=-1;
    }
    else if(i==3) {
        NN[0]=-1;
        NN[1]=0;
    }
    else if(i==4) {
        NN[0]=1;
        NN[1]=0;
    }
    else if(i==5) {
        NN[0]=-1;
        NN[1]=1;
    }
    else if(i==6) {
        NN[0]=0;      NN[1]=1;
    }
    else {
        NN[0]=1;
        NN[1]=1;
    }
    return NN;
}

// Matlab main function //

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{   int i, j, m, n, max_label, temp_label, t;
    int *min_area;
    priority_queue<couple, vector<couple>, C> PQ, temp;
    
    double REALMAX = 1000.0;
    
    double *data = mxGetPr(prhs[0]);
    int h = mxGetM(prhs[0]);
    int w = mxGetN(prhs[0]);
    
    double *tlambda = mxGetPr(prhs[1]);
    int lambda = tlambda[0];
    
    plhs[0] = mxCreateDoubleMatrix(h, w, mxREAL);
    double *out = mxGetPr(plhs[0]);
    
    // Initialization
    vi im_lab(h*w);
    for(j=0;j<h;j++)
        for( i=0;i<w;i++){
        im_lab.at(i+j*w).x=i;
        im_lab.at(i+j*w).y=j;
        im_lab.at(i+j*w).gray=data[i*h+j];
        im_lab.at(i+j*w).label=0;
        }
    
    // Run the area filtering
    int lambda_tmp=2;
    while(lambda_tmp<lambda) {
        //Labelization of the Image & Initialisation of the parameter
        im_lab=labelization(im_lab, h, w); //labelize flat zone
        max_label=find_max_label(im_lab);//find the number of label
        ve area_tab=compute_area(im_lab, max_label);//compute the corresponding area for each label
        
        //label the seed
        
        for(i=0;i<im_lab.size();i++)
            if(area_tab[im_lab[i].label-1]<lambda_tmp) {
            area_tab.at(im_lab[i].label-1)=0;
            im_lab.at(i).label=0;
            temp_label++;
            }
        
        //Initialisation of the Priority Queue
        for(j=0;j<h;j++)
            for(i=0;i<w;i++)
                if(im_lab[i+j*w].label>0) {
            for(m=j-1;m<j+2;m++)
                for(n=i-1;n<i+2;n++)
                    //scan for unlabeled neighbors
                    if((0<=m)&&(m<h)&&(0<=n)&&(n<w)&&(im_lab[n+m*w].label==0)) {
                PQ.push(couple(REALMAX-abs(im_lab[i+j*w].gray -im_lab[n+m*w].gray), 0, im_lab[n+m*w].x, im_lab[n+m*w].y));
                im_lab.at(n+m*w).label=-1;
                    }
                }
        
        //Region Growing
        while(!PQ.empty()) {
            i=PQ.top().x;
            j=PQ.top().y;
            PQ.pop();
            
            //scan neighbors
            vd neighbors(8);
            t=0;
            for(m=j-1;m<j+2;m++)
                for(n=i-1;n<i+2;n++) {
                if((0<=m)&&(m<h)&&(0<=n)&&(n<w)&&(n!=i||m!=j)&&(im_lab[n+w*m].label>0))
                    neighbors.at(t)=REALMAX-abs(im_lab[i+j*w].gray-im_lab[n+m*w].gray);
                t++;
                if(n==i && m==j)
                    t--;
                }
            
            //find the closet labelled pixel in term of gray value.
            int *nn=convert_indice_2_neighbors(indice_max_value(neighbors));
            
            //set the pixel & the label value to the corresponding value.
            im_lab.at(i+j*w).gray=im_lab[(i+nn[0])+(j+nn[1])*w].gray;
            im_lab.at(i+j*w).label=im_lab[(i+nn[0])+(j+nn[1])*w].label;
            
            //add the neighbors to the PQ list
            for(m=j-1;m<j+2;m++)
                for(n=i-1;n<i+2;n++) {
                if((0<=m)&&(m<h)&&(0<=n)&&(n<w)&&(n!=i||m!=j)&&(im_lab[n+w*m].label==0))//add unlabeled pixel that are not in the queue
                {
                    PQ.push(couple(REALMAX-abs(im_lab[i+j*w].gray -im_lab[n+m*w].gray), 0, im_lab[n+m*w].x, im_lab[n+m*w].y));
                    im_lab.at(n+m*w).label=-1;
                }
                }
            delete nn;
            neighbors.clear();
        }
        lambda_tmp++;
    }
    
    for(j=0;j<h;j++)
        for( i=0;i<w;i++){
        out[i*h+j]=im_lab.at(i+j*w).gray;
        }
}